USER_PROXY_DESCRIPTION = """
The human user who gave the original task.
The human user has access to the same browser as the websurfer. However, they do not have the ability to write or execute code.
In case where the task requires further clarifying information, the user can be asked to clarify the task.
In case where you are stuck and unable to make progress on completing the task, you can ask the user for help.
Make sure to do your best to complete the task with other agents before asking the user for help.
The user can help you complete captchas or other tasks that require human intervention if necessary.
"""
