"use strict";(self.webpackChunkDr_Sai=self.webpackChunkDr_Sai||[]).push([[577],{74749:function(e,a,t){t.r(a);var s=t(96540),n=t(64810),o=t(92744);a.default=()=>{const{setUser:e}=s.useContext(o.v);return s.useEffect((()=>{const a=new URLSearchParams(window.location.search),t=a.get("token"),s=a.get("username");t&&s?(localStorage.setItem("token",t),localStorage.setItem("username",s),localStorage.setItem("user_email",s),localStorage.setItem("user_name",s),localStorage.removeItem("drsai-mode-config"),e({name:s,email:s,username:s}),(0,n.navigate)("/")):(0,n.navigate)("/sso-login")}),[]),s.createElement("div",null,"正在登录，请稍候...")}}}]);
//# sourceMappingURL=component---src-pages-auth-tsx-8b49e02b0dd162a405e6.js.map