
from .input_func import AsyncInputFunc, InputFuncType, InputRequestType, SyncInputFunc
from .version import __version__

ABOUT = "Magentic-UI - A web browsing assistant."
__all__ = [
    "AsyncInputFunc",
    "InputFuncType",
    "InputRequestType",
    "SyncInputFunc",
    "__version__",
]
