from .web_surfer import WebSurfer, WebSurferCUA
from ._coder import CoderAgent
from ._user_proxy import USER_PROXY_DESCRIPTION
from .file_surfer import FileSurfer
from .drsai_agents import MagenticAgent
__all__ = [
    "WebSurfer",
    "WebSurferCUA",
    "CoderAgent",
    "USER_PROXY_DESCRIPTION",
    "FileSurfer",
    "MagenticAgent",
]
